#include "scene.h"
#include <iostream>
#include <string>
#include "node.h"
#include "transformnode.h"
#include "translatenode.h"
#include "rotatenode.h"
#include "geometrynode.h"
#include "boxnode.h"
#include "cameranode.h"
#include "Matrix.h"

Scene::Scene(QObject* parent) : QObject(parent)
{
    Node* r = new Node("Root");

    Node* r1 = new RotateNode(45.0, 0,1,0, "RRot");
    r->addChild(r1);
    Node* t1 = new TranslateNode(-2.0,0,0, "TRot");
    r1->addChild(t1);
    Node* g1 = new BoxNode("Rot",1.0,0,0);
    t1->addChild(g1);

    Node* t2 = new TranslateNode(2.0,0,0, "TBlau");
    r->addChild(t2);
    Node* r2 = new RotateNode(45.0, 0,1,0, "RBlau");
    t2->addChild(r2);
    Node* g2 = new BoxNode("Blau",0,0,1.0);
    r2->addChild(g2);

    Node* t3 = new TranslateNode(0,0,4.0, "TCamera");
    r->addChild(t3);
    Node* r3 = new RotateNode(0.0, 0,1,0, "RCamera");
    t3->addChild(r3);
    CameraNode* c = new CameraNode("Camera1");
    r3->addChild(c);

    this->red = g1;
    this->blue = g2;
    root = r;
    root->print();
    camera = c;
    wasFocussed = false;
}

GLdouble* Scene::getNodeTransform(Node* node){
    glLoadIdentity();
    Node* curr = node;
    while (curr->getParent() != NULL && curr->getParent() != root){
        curr = curr->getParent();
//        std::cout << "Travelling to " << curr->getName() << "(" << curr->getType() << ")" << std::endl;
        if (curr->getType().compare("TransformNode") == 0 ){
//            std::cout << "Inverting TransformNode " << curr->getName() << std::endl;
            ((TransformNode*) curr)->applyInverse();
        }
    }
    GLdouble mat[16];
    glGetDoublev(GL_MODELVIEW_MATRIX, mat);
    GLdouble* pos = new GLdouble[3];
    pos[0] = - mat[12];
    pos[1] = - mat[13];
    pos[2] = - mat[14];
    return pos;
}

static void setCameraTransform(Node* node, Node* root)
{
    glLoadIdentity();
    Node* curr = node;
    while (curr->getParent() != NULL && curr->getParent() != root){
        curr = curr->getParent();
        if (curr->getType() == "TransformNode"){
            ((TransformNode*) curr)->applyInverse();
        }
    }
}

void Scene::render()
{
    glLoadIdentity();
    setCameraTransform(camera, root);

    root->apply();
}

Scene::~Scene()
{
    delete root;
}

void Scene::replaceCamRotationNode(Node* newRotateNode)
{
     Node* oldRotateNode = camera->getParent();
     Node* parent = oldRotateNode->getParent();
     parent->removeChild(oldRotateNode);
     oldRotateNode->removeChild(camera); //remove child camera else camera gets deleted once parent is deleted
     delete oldRotateNode;

     parent->addChild(newRotateNode);
     newRotateNode->addChild(camera);
}

void Scene::rotateCamX(int val)
{
    rotX = val;
    if (wasFocussed)
        return;

    Matrix m;

    m.rotate(rotZ, Matrix::Z);
    m.rotate(rotY, Matrix::Y);
    m.rotate(rotX, Matrix::X);

    RotateNode* newRotateNode = new RotateNode(m.get());
    replaceCamRotationNode(newRotateNode);
}

void Scene::rotateCamY(int val)
{
    rotY = val;
    if (wasFocussed)
        return;

    Matrix m;

    m.rotate(rotZ, Matrix::Z);
    m.rotate(rotY, Matrix::Y);
    m.rotate(rotX, Matrix::X);

    RotateNode* newRotateNode = new RotateNode(m.get());
    replaceCamRotationNode(newRotateNode);
}

void Scene::rotateCamZ(int val)
{
    rotZ = val;
    if (wasFocussed)
        return;

    Matrix m;

    m.rotate(rotZ, Matrix::Z);
    m.rotate(rotY, Matrix::Y);   
    m.rotate(rotX, Matrix::X);

    RotateNode* newRotateNode = new RotateNode(m.get());
    replaceCamRotationNode(newRotateNode);
}

//Applies all Transforms from Root node to the current Node and saves result in combined matrix
Matrix Scene::forwardApply(Node* originNode) {
    std::vector<GLdouble*> nodeTransforms;
    Node* curr = originNode;
    while (curr->getParent() != NULL && curr->getParent() != root)
    {
        if (curr->getParent()->getType() == "TransformNode")
        {
            TransformNode* transformNode = static_cast<TransformNode*>(curr->getParent());
            nodeTransforms.push_back(transformNode->get());
        }
        curr = curr->getParent();
    }
    Matrix last;
    for (std::vector<GLdouble*>::reverse_iterator it = nodeTransforms.rbegin(); it != nodeTransforms.rend(); it++)
    {
        Matrix matrix(*it);
        last.mult(matrix);
    }
    return last;
}

//calculates the direction vector between the camera node and
//the selected node in world coordinates and generates
//the rotation matrix based on this vector.
//the camera rotation node is replaced with a new rotation node
//based on that rotation matrix
void Scene::focus(const std::string& nodeName)
{
    Node* target = root->findChildNodeByName(nodeName);
    if (target)
    {
        Matrix targetTransform = forwardApply(target);
        Vector targetPos = targetTransform.translation();

        Matrix cameraTransform = forwardApply(camera);
        Vector cameraPos = cameraTransform.translation();

        Vector viewDirection = targetPos - cameraPos;

        Matrix m = Matrix::fromViewDir(viewDirection);
        RotateNode* newRot = new RotateNode(m.get());
        replaceCamRotationNode(newRot);

        wasFocussed = true;
        float x,y,z;
        m.euler(&x, &y, &z);
        emit rotationXChanged(x);
        emit rotationYChanged(y);
        emit rotationZChanged(z);
        wasFocussed = false;
    }
}
